import warnings
from typing import Tuple
import torch
import torch.nn.functional as F
from asteroid_filterbanks import STFTFB
from asteroid_filterbanks.scripting import script_if_tracing


class TorchSTFTFB(STFTFB):
    """Equivalent to :function:`torch.stft` and :function:`torch.istft` using
    `asteroid-filterbanks`.

    Args:
        *args: Passed to `STFTFB`.
        center (bool): Whether to center the each frame. (pad left and right).
        pad_mode (str): How to pad if `center` is True. Only used for the STFT.
        normalize: Unsupported yet.
        **kwargs: Passed to `STFTFB`.

    See :class:`STFTFB`, :function:`torch.stft` and :function:`torch.istft` for
    more detail.

    .. warning::
        In order to obtain perfect reconstruction, `torch.istft` divides the wav
        output by the overlap-added square window (WSOLA) (assumed the same for
         analysis and synthesis). This might cause problems for windows containing
         zeros (and several common windows do). When using `center=True`, the leading
        and trailing zeros of the WSOLA will be discarded, virtually solving this
        problem.
        We intentionally match this behavior, to provide code that behaves exactly
        like `torch.istft`. But whereas `torch.istft` raises an error when the
        WSOLA contains zeros, we only raise a warning.
        As the code is rigorously tested, if your code worked before with
        `torch.istft`, it won't break with TorchSTFTFB.

    .. note::
        When no `length` argument is given to `torch.istft`, the padding is
        aggressively removed. In this implementation, the whole signal is passed
        back to the user.
    """

    def __init__(
        self,
        n_filters,
        kernel_size,
        stride=None,
        window=None,
        center=True,
        pad_mode="reflect",
        normalized=False,
        onesided=True,
        sample_rate=8000.0,
        **kwargs,
    ):
        if n_filters != kernel_size:
            raise NotImplementedError(
                "Cannot set `n_filters!=kernel_size` in TorchSTFTFB, untested."
            )
        super().__init__(
            n_filters, kernel_size, stride=stride, window=window, sample_rate=sample_rate, **kwargs
        )
        self.center = center
        self.pad_mode = pad_mode
        if normalized:
            raise NotImplementedError
        if not onesided:
            raise NotImplementedError
        self.normalize = normalized

    @classmethod
    def from_torch_args(
        cls,
        n_fft,
        hop_length=None,
        win_length=None,
        window=None,
        center=True,
        pad_mode="reflect",
        normalized=False,
        onesided=True,
        sample_rate=8000.0,
        **kwargs,
    ):
        return cls(
            n_filters=n_fft,
            kernel_size=win_length,
            stride=hop_length,
            window=window,
            center=center,
            pad_mode=pad_mode,
            normalized=normalized,
            onesided=onesided,
            sample_rate=sample_rate,
            **kwargs,
        )

    def pre_analysis(self, wav):
        """Centers the frames if `center` is True."""
        if not self.center:
            return wav
        pad_shape = (self.kernel_size // 2, self.kernel_size // 2)
        wav = pad_all_shapes(wav, pad_shape=pad_shape, mode=self.pad_mode)
        return wav

    def post_analysis(self, spec):
        """Correct the scale to match torch.stft."""
        spec = _restore_freqs_an(spec, n_filters=self.n_filters)
        return spec * 0.5 * (self.kernel_size * self.n_filters / self.stride) ** 0.5

    def pre_synthesis(self, spec):
        """Correct the scale to match what's expected by torch.istft."""
        spec = spec.clone()
        spec = _restore_freqs_syn(spec, n_filters=self.n_filters)
        return spec / ((self.stride * self.n_filters / self.kernel_size) ** 0.5)

    def post_synthesis(self, wav):
        """Perform OLA on the waveforms and divide by the squared window OLA.
        Assumes that the window is the same in the STFT and iSTFT.
        """
        wav_out = ola_with_wdiv(
            wav,
            window=self.torch_window,
            kernel_size=self.kernel_size,
            stride=self.stride,
            center=self.center,
        )
        return wav_out


@script_if_tracing
def _restore_freqs_an(spec, n_filters: int):
    spec[..., 0, :] *= 2 ** 0.5
    spec[..., n_filters // 2, :] *= 2 ** 0.5
    return spec


@script_if_tracing
def _restore_freqs_syn(spec, n_filters: int):
    spec[..., 0, :] /= 2 ** 0.5
    spec[..., n_filters // 2, :] /= 2 ** 0.5
    return spec


@script_if_tracing
def ola_with_wdiv(wav, window, kernel_size: int, stride: int, center: bool = True) -> torch.Tensor:
    n_frame = 1 + (wav.shape[-1] - kernel_size) // stride
    wsq_ola = square_ola(
        window,
        kernel_size=kernel_size,
        stride=stride,
        n_frame=n_frame,
    )
    view_shape = [1 for _ in wav.shape[:-1]] + [-1]
    wsq_ola = wsq_ola.view(view_shape)

    start = kernel_size // 2 if center else 0
    wav = wav[..., start:]
    wsq_ola = wsq_ola[..., start:].expand_as(wav)

    min_mask = wsq_ola.abs() < 1e-11
    if min_mask.any() and not torch.jit.is_scripting():
        # Warning instead of error. Might be trimmed afterward.
        warnings.warn(
            f"Minimum NOLA should be above 1e-11, Found {wsq_ola.abs().min()}. "
            f"Dividind only where possible.",
            RuntimeWarning,
        )
    wav[~min_mask] = wav[~min_mask] / wsq_ola[~min_mask]
    return wav


def square_ola(window: torch.Tensor, kernel_size: int, stride: int, n_frame: int) -> torch.Tensor:
    window_sq = window.pow(2).view(1, -1, 1).repeat(1, 1, n_frame)
    return torch.nn.functional.fold(
        window_sq, (1, (n_frame - 1) * stride + kernel_size), (1, kernel_size), stride=(1, stride)
    ).squeeze(2)


@script_if_tracing
def pad_all_shapes(
    x: torch.Tensor, pad_shape: Tuple[int, int], mode: str = "reflect"
) -> torch.Tensor:
    if x.ndim == 1:
        return F.pad(x[None, None], pad=pad_shape, mode=mode).squeeze(0).squeeze(0)
    if x.ndim == 2:
        return F.pad(x[None], pad=pad_shape, mode=mode).squeeze(0)
    if x.ndim == 3:
        return F.pad(x, pad=pad_shape, mode=mode)
    pad_shape = (pad_shape[0],) + (0,) * (x.ndim - 1)
    return F.pad(x, pad=pad_shape, mode=mode)



import warnings
from functools import wraps


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.

    By default, python will not show deprecation warnings, so this class
    can be used when a very visible warning is helpful, for example because
    the usage is most likely a user bug.

    """

    # Taken from numpy


def mark_deprecated(message, version=None):
    """Decorator to add deprecation message.

    Args:
        message: Migration steps to be given to users.
    """

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            from_what = "a future release" if version is None else f"asteroid v{version}"
            warn_message = (
                f"{func.__module__}.{func.__name__} has been deprecated "
                f"and will be removed from {from_what}. "
                f"{message}"
            )
            warnings.warn(warn_message, VisibleDeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped

    return decorator



import torch
from . import STFTFB
from . import transforms


class MelGramFB(STFTFB):
    """Mel magnitude spectrogram filterbank.

    Args:
        n_filters (int): Number of filters. Determines the length of the STFT
            filters before windowing.
        kernel_size (int): Length of the filters (i.e the window).
        stride (int, optional): Stride of the convolution (hop size). If None
            (default), set to ``kernel_size // 2``.
        window (:class:`numpy.ndarray`, optional): If None, defaults to
            ``np.sqrt(np.hanning())``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.
        n_mels (int): Number of mel bands.
        fmin (float): Minimum frequency of the mel filters.
        fmax (float): Maximum frequency of the mel filters. Defaults to sample_rate//2.
        norm (str): Mel normalization {None, 'slaney', or number}.
            See `librosa.filters.mel`
        **kwargs:
    """

    def __init__(
        self,
        n_filters,
        kernel_size,
        stride=None,
        window=None,
        sample_rate=8000.0,
        n_mels=40,
        fmin=0.0,
        fmax=None,
        norm="slaney",
        **kwargs,
    ):

        self.n_mels = n_mels
        self.fmin = fmin
        self.fmax = fmax
        self.norm = norm
        super().__init__(
            n_filters=n_filters,
            kernel_size=kernel_size,
            stride=stride,
            window=window,
            sample_rate=sample_rate,
            **kwargs,
        )
        self.mel_scale = MelScale(
            n_filters, sample_rate=sample_rate, n_mels=n_mels, fmin=fmin, fmax=fmax, norm=norm
        )
        self.n_feats_out = n_mels

    def post_analysis(self, spec: torch.Tensor):
        return self.mel_scale(spec)

    def get_config(self):
        conf = dict(fmin=self.fmin, fmax=self.fmax, n_mels=self.n_mels, norm=self.norm)
        return {**super().get_config(), **conf}


class MelScale(torch.nn.Module):
    """Mel-scale filterbank matrix.

    Args:
        n_filters (int): Number of filters. Determines the length of the STFT
            filters before windowing.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.
        n_mels (int): Number of mel bands.
        fmin (float): Minimum frequency of the mel filters.
        fmax (float): Maximum frequency of the mel filters. Defaults to sample_rate//2.
        norm (str): Mel normalization {None, 'slaney', or number}.
            See `librosa.filters.mel`
    """

    def __init__(
        self,
        n_filters,
        sample_rate=8000.0,
        n_mels=40,
        fmin=0.0,
        fmax=None,
        norm="slaney",
    ):
        from librosa.filters import mel

        super().__init__()
        fb_mat = mel(
            sr=sample_rate, n_fft=n_filters, fmin=fmin, fmax=fmax, n_mels=n_mels, norm=norm
        )
        self.register_buffer("fb_mat", torch.from_numpy(fb_mat).unsqueeze(0))

    def forward(self, spec: torch.Tensor):
        mag_spec = transforms.mag(spec, dim=-2)
        mel_spec = torch.matmul(self.fb_mat, mag_spec)
        return mel_spec



import numpy as np
import torch
from .enc_dec import Filterbank


class MultiphaseGammatoneFB(Filterbank):
    r"""Multi-Phase Gammatone Filterbank as described in [1].

    Please cite [1] whenever using this.

    `Original code repository: <https://github.com/sp-uhh/mp-gtf>`_

    Args:
        n_filters (int): Number of filters.
        kernel_size (int): Length of the filters.
        sample_rate (float, optional): The sample rate (used for initialization).
        stride (int, optional): Stride of the convolution. If None (default),
            set to ``kernel_size // 2``.

    References
        [1] David Ditter, Timo Gerkmann, "A Multi-Phase Gammatone Filterbank for
        Speech Separation via TasNet", ICASSP 2020
        Available: https://ieeexplore.ieee.org/document/9053602/
    """

    def __init__(self, n_filters=128, kernel_size=16, sample_rate=8000.0, stride=None, **kwargs):
        super().__init__(n_filters, kernel_size, stride=stride, sample_rate=sample_rate)
        self.n_feats_out = n_filters
        length_in_seconds = kernel_size / sample_rate
        mpgtf = generate_mpgtf(sample_rate, length_in_seconds, n_filters)
        filters = torch.from_numpy(mpgtf).unsqueeze(1).float()
        self.register_buffer("_filters", filters)

    def filters(self):
        return self._filters


def generate_mpgtf(samplerate_hz, len_sec, n_filters):
    # Set parameters
    center_freq_hz_min = 100
    n_center_freqs = 24
    len_sample = int(np.floor(samplerate_hz * len_sec))

    # Initialize variables
    index = 0
    filterbank = np.zeros((n_filters, len_sample))
    current_center_freq_hz = center_freq_hz_min

    # Determine number of phase shifts per center frequency
    phase_pair_count = (np.ones(n_center_freqs) * np.floor(n_filters / 2 / n_center_freqs)).astype(
        int
    )
    remaining_phase_pairs = ((n_filters - np.sum(phase_pair_count) * 2) / 2).astype(int)
    if remaining_phase_pairs > 0:
        phase_pair_count[:remaining_phase_pairs] = phase_pair_count[:remaining_phase_pairs] + 1

    # Generate all filters for each center frequencies
    for i in range(n_center_freqs):
        # Generate all filters for all phase shifts
        for phase_index in range(phase_pair_count[i]):
            # First half of filtes: phase_shifts in [0,pi)
            current_phase_shift = np.float(phase_index) / phase_pair_count[i] * np.pi
            filterbank[index, :] = gammatone_impulse_response(
                samplerate_hz,
                len_sec,
                current_center_freq_hz,
                current_phase_shift,
            )
            index = index + 1

        # Second half of filters: phase_shifts in [pi, 2*pi)
        filterbank[index : index + phase_pair_count[i], :] = -filterbank[
            index - phase_pair_count[i] : index, :
        ]

        # Prepare for next center frequency
        index = index + phase_pair_count[i]
        current_center_freq_hz = erb_scale_2_freq_hz(
            freq_hz_2_erb_scale(current_center_freq_hz) + 1
        )

    filterbank = normalize_filters(filterbank)
    return filterbank


def gammatone_impulse_response(samplerate_hz, len_sec, center_freq_hz, phase_shift):
    """ Generate single parametrized gammatone filter """
    p = 2  # filter order
    erb = 24.7 + 0.108 * center_freq_hz  # equivalent rectangular bandwidth
    divisor = (np.pi * np.math.factorial(2 * p - 2) * np.power(2, float(-(2 * p - 2)))) / np.square(
        np.math.factorial(p - 1)
    )
    b = erb / divisor  # bandwidth parameter
    a = 1.0  # amplitude. This is varied later by the normalization process.
    len_sample = int(np.floor(samplerate_hz * len_sec))
    t = np.linspace(1.0 / samplerate_hz, len_sec, len_sample)
    gammatone_ir = (
        a
        * np.power(t, p - 1)
        * np.exp(-2 * np.pi * b * t)
        * np.cos(2 * np.pi * center_freq_hz * t + phase_shift)
    )
    return gammatone_ir


def erb_scale_2_freq_hz(freq_erb):
    """ Convert frequency on ERB scale to frequency in Hertz """
    freq_hz = (np.exp(freq_erb / 9.265) - 1) * 24.7 * 9.265
    return freq_hz


def freq_hz_2_erb_scale(freq_hz):
    """ Convert frequency in Hertz to frequency on ERB scale """
    freq_erb = 9.265 * np.log(1 + freq_hz / (24.7 * 9.265))
    return freq_erb


def normalize_filters(filterbank):
    """Normalizes a filterbank such that all filters
    have the same root mean square (RMS)."""
    rms_per_filter = np.sqrt(np.mean(np.square(filterbank), axis=1))
    rms_normalization_values = 1.0 / (rms_per_filter / np.amax(rms_per_filter))
    normalized_filterbank = filterbank * rms_normalization_values[:, np.newaxis]
    return normalized_filterbank



import torch
import math
from typing import Optional, List

from . import Encoder, Decoder, STFTFB
from .stft_fb import perfect_synthesis_window
from . import transforms


def griffin_lim(mag_specgram, stft_enc, angles=None, istft_dec=None, n_iter=6, momentum=0.9):
    """Estimates matching phase from magnitude spectogram using the
    'fast' Griffin Lim algorithm [1].

    Args:
        mag_specgram (torch.Tensor): (any, dim, ension, freq, frames) as
            returned by `Encoder(STFTFB)`, the magnitude spectrogram to be
            inverted.
        stft_enc (Encoder[STFTFB]): The `Encoder(STFTFB())` object that was
            used to compute the input `mag_spec`.
        angles (None or Tensor): Angles to use to initialize the algorithm.
            If None (default), angles are init with uniform ditribution.
        istft_dec (None or Decoder[STFTFB]): Optional Decoder to use to get
            back to the time domain. If None (default), a perfect
            reconstruction Decoder is built from `stft_enc`.
        n_iter (int): Number of griffin-lim iterations to run.
        momentum (float): The momentum of fast Griffin-Lim. Original
            Griffin-Lim is obtained for momentum=0.

    Returns:
        torch.Tensor: estimated waveforms of shape (any, dim, ension, time).

    Examples
        >>> stft = Encoder(STFTFB(n_filters=256, kernel_size=256, stride=128))
        >>> wav = torch.randn(2, 1, 8000)
        >>> spec = stft(wav)
        >>> masked_spec = spec * torch.sigmoid(torch.randn_like(spec))
        >>> mag = transforms.mag(masked_spec, -2)
        >>> est_wav = griffin_lim(mag, stft, n_iter=32)

    References
        [1] Perraudin et al. "A fast Griffin-Lim algorithm," WASPAA 2013.

        [2] D. W. Griffin and J. S. Lim:  "Signal estimation from modified
        short-time Fourier transform," ASSP 1984.

    """
    # We can create perfect iSTFT from STFT Encoder
    if istft_dec is None:
        # Compute window for perfect resynthesis
        syn_win = perfect_synthesis_window(stft_enc.filterbank.window, stft_enc.stride)
        istft_dec = Decoder(STFTFB(**stft_enc.get_config(), window=syn_win))

    # If no intitial phase is provided initialize uniformly
    if angles is None:
        angles = 2 * math.pi * torch.rand_like(mag_specgram, device=mag_specgram.device)
    else:
        angles = angles.view(*mag_specgram.shape)

    # Initialize rebuilt (useful to use momentum)
    rebuilt = 0.0
    for _ in range(n_iter):
        prev_built = rebuilt
        # Go to the time domain
        complex_specgram = transforms.from_magphase(mag_specgram, angles)
        waveform = istft_dec(complex_specgram)
        # And back to TF domain
        rebuilt = stft_enc(waveform)
        # Update phase estimates (with momentum)
        diff = rebuilt - momentum / (1 + momentum) * prev_built
        angles = transforms.angle(diff)

    final_complex_spec = transforms.from_magphase(mag_specgram, angles)
    return istft_dec(final_complex_spec)


def misi(
    mixture_wav,
    mag_specgrams,
    stft_enc,
    angles=None,
    istft_dec=None,
    n_iter=6,
    momentum=0.0,
    src_weights=None,
    dim=1,
):
    """Jointly estimates matching phase from magnitude spectograms using the
    Multiple Input Spectrogram Inversion (MISI) algorithm [1].

    Args:
        mixture_wav (torch.Tensor): (batch, time)
        mag_specgrams (torch.Tensor): (batch, n_src, freq, frames) as
            returned by `Encoder(STFTFB)`, the magnitude spectrograms to be
            jointly inverted using MISI (modified or not).
        stft_enc (Encoder[STFTFB]): The `Encoder(STFTFB())` object that was
            used to compute the input `mag_spec`.
        angles (None or Tensor): Angles to use to initialize the algorithm.
            If None (default), angles are init with uniform ditribution.
        istft_dec (None or Decoder[STFTFB]): Optional Decoder to use to get
            back to the time domain. If None (default), a perfect
            reconstruction Decoder is built from `stft_enc`.
        n_iter (int): Number of MISI iterations to run.
        momentum (float): Momentum on updates (this argument comes from
            GriffinLim). Defaults to 0 as it was never proposed anywhere.
        src_weights (None or torch.Tensor): Consistency weight for each source.
            Shape needs to be broadcastable to `istft_dec(mag_specgrams)`.
            We make sure that the weights sum up to 1 along dim `dim`.
            If `src_weights` is None, compute them based on relative power.
        dim (int): Axis which contains the sources in `mag_specgrams`.
            Used for consistency constraint.

    Returns:
        torch.Tensor: estimated waveforms of shape (batch, n_src, time).

    Examples
        >>> stft = Encoder(STFTFB(n_filters=256, kernel_size=256, stride=128))
        >>> wav = torch.randn(2, 3, 8000)
        >>> specs = stft(wav)
        >>> masked_specs = specs * torch.sigmoid(torch.randn_like(specs))
        >>> mag = transforms.mag(masked_specs, -2)
        >>> est_wav = misi(wav.sum(1), mag, stft, n_iter=32)

    References
        [1] Gunawan and Sen, "Iterative Phase Estimation for the Synthesis of
        Separated Sources From Single-Channel Mixtures," in IEEE Signal
        Processing Letters, 2010.

        [2] Wang, LeRoux et al. “End-to-End Speech Separation with Unfolded
        Iterative Phase Reconstruction.” Interspeech 2018 (2018)
    """
    # We can create perfect iSTFT from STFT Encoder
    if istft_dec is None:
        # Compute window for perfect resynthesis
        syn_win = perfect_synthesis_window(stft_enc.filterbank.window, stft_enc.stride)
        istft_dec = Decoder(STFTFB(**stft_enc.get_config(), window=syn_win))

    # If no intitial phase is provided initialize uniformly
    if angles is None:
        angles = 2 * math.pi * torch.rand_like(mag_specgrams, device=mag_specgrams.device)
    # wav_dim is used in mixture_consistency.
    # Transform spec src dim to wav src dim for positive and negative dim
    wav_dim = dim if dim >= 0 else dim + 1

    # We forward/backward the mixture through STFT to have matching shapes
    # with the input spectrograms as well as  account for potential modulations
    # if the window were not chosen to enable perfect reconstruction.
    mixture_wav = istft_dec(stft_enc(mixture_wav))

    # Initialize rebuilt (useful to use momentum)
    rebuilt = 0.0
    for _ in range(n_iter):
        prev_built = rebuilt
        # Go to the time domain
        complex_specgram = transforms.from_magphase(mag_specgrams, angles)
        wavs = istft_dec(complex_specgram)
        # Make wavs sum up to the mixture
        consistent_wavs = _mixture_consistency(
            mixture_wav, wavs, src_weights=src_weights, dim=wav_dim
        )
        # Back to TF domain
        rebuilt = stft_enc(consistent_wavs)
        # Update phase estimates (with momentum). Keep the momentum here
        # in case. Was shown useful in GF, might be here. We'll see.
        diff = rebuilt - momentum / (1 + momentum) * prev_built
        angles = transforms.angle(diff)
    # Final source estimates
    final_complex_spec = transforms.from_magphase(mag_specgrams, angles)
    return istft_dec(final_complex_spec)


def _mixture_consistency(
    mixture: torch.Tensor,
    est_sources: torch.Tensor,
    src_weights: Optional[torch.Tensor] = None,
    dim: int = 1,
) -> torch.Tensor:
    """Applies mixture consistency to a tensor of estimated sources.

    Args:
        mixture (torch.Tensor): Mixture waveform or TF representation.
        est_sources (torch.Tensor): Estimated sources waveforms or TF
            representations.
        src_weights (torch.Tensor): Consistency weight for each source.
            Shape needs to be broadcastable to `est_source`.
            We make sure that the weights sum up to 1 along dim `dim`.
            If `src_weights` is None, compute them based on relative power.
        dim (int): Axis which contains the sources in `est_sources`.

    Returns:
        torch.Tensor with same shape as `est_sources`, after applying mixture
        consistency.

    .. note::
        This method can be used only in 'complete' separation tasks, otherwise
        the residual error will contain unwanted sources. For example, this
        won't work with the task `sep_noisy` from WHAM.

    Examples
        >>> # Works on waveforms
        >>> mix = torch.randn(10, 16000)
        >>> est_sources = torch.randn(10, 2, 16000)
        >>> new_est_sources = mixture_consistency(mix, est_sources, dim=1)
        >>> # Also works on spectrograms
        >>> mix = torch.randn(10, 514, 400)
        >>> est_sources = torch.randn(10, 2, 514, 400)
        >>> new_est_sources = mixture_consistency(mix, est_sources, dim=1)

    References
        Scott Wisdom, John R Hershey, Kevin Wilson, Jeremy Thorpe, Michael
        Chinen, Brian Patton, and Rif A Saurous. "Differentiable consistency
        constraints for improved deep speech enhancement", ICASSP 2019.
    """
    # If the source weights are not specified, the weights are the relative
    # power of each source to the sum. w_i = P_i / (P_all), P for power.
    if src_weights is None:
        all_dims: List[int] = torch.arange(est_sources.ndim).tolist()
        all_dims.pop(dim)  # Remove source axis
        all_dims.pop(0)  # Remove batch axis
        src_weights = torch.mean(est_sources ** 2, dim=all_dims, keepdim=True)
    # Make sure that the weights sum up to 1
    norm_weights = torch.sum(src_weights, dim=dim, keepdim=True) + 1e-8
    src_weights = src_weights / norm_weights

    # Compute residual mix - sum(est_sources)
    if mixture.ndim == est_sources.ndim - 1:
        # mixture (batch, *), est_sources (batch, n_src, *)
        residual = (mixture - est_sources.sum(dim=dim)).unsqueeze(dim)
    elif mixture.ndim == est_sources.ndim:
        # mixture (batch, 1, *), est_sources (batch, n_src, *)
        residual = mixture - est_sources.sum(dim=dim, keepdim=True)
    else:
        n, m = est_sources.ndim, mixture.ndim
        raise RuntimeError(
            f"The size of the mixture tensor should match the "
            f"size of the est_sources tensor. Expected mixture"
            f"tensor to have {n} or {n-1} dimension, found {m}."
        )
    # Compute remove
    new_sources = est_sources + src_weights * residual
    return new_sources



import numpy as np
import torch
import torch.nn as nn
from .enc_dec import Filterbank


class ParamSincFB(Filterbank):
    """Extension of the parameterized filterbank from [1] proposed in [2].
    Modified and extended from from `<https://github.com/mravanelli/SincNet>`__

    Args:
        n_filters (int): Number of filters. Half of `n_filters` (the real
            parts) will have parameters, the other half will correspond to the
            imaginary parts. `n_filters` should be even.
        kernel_size (int): Length of the filters.
        stride (int, optional): Stride of the convolution. If None (default),
            set to ``kernel_size // 2``.
        sample_rate (float, optional): The sample rate (used for initialization).
        min_low_hz (int, optional): Lowest low frequency allowed (Hz).
        min_band_hz (int, optional): Lowest band frequency allowed (Hz).

    Attributes:
        n_feats_out (int): Number of output filters.

    References
        [1] : "Speaker Recognition from raw waveform with SincNet". SLT 2018.
        Mirco Ravanelli, Yoshua Bengio.  https://arxiv.org/abs/1808.00158

        [2] : "Filterbank design for end-to-end speech separation". ICASSP 2020.
        Manuel Pariente, Samuele Cornell, Antoine Deleforge, Emmanuel Vincent.
        https://arxiv.org/abs/1910.10400
    """

    def __init__(
        self,
        n_filters,
        kernel_size,
        stride=None,
        sample_rate=16000.0,
        min_low_hz=50,
        min_band_hz=50,
        **kwargs,
    ):
        if kernel_size % 2 == 0:
            print(
                "Received kernel_size={}, force ".format(kernel_size)
                + "kernel_size={} so filters are odd".format(kernel_size + 1)
            )
            kernel_size += 1
        super().__init__(n_filters, kernel_size, stride=stride, sample_rate=sample_rate)
        self.min_low_hz, self.min_band_hz = min_low_hz, min_band_hz

        self.half_kernel = self.kernel_size // 2
        self.cutoff = int(n_filters // 2)
        self.n_feats_out = 2 * self.cutoff
        self._initialize_filters()
        if n_filters % 2 != 0:
            print(
                "If the number of filters `n_filters` is odd, the "
                "output size of the layer will be `n_filters - 1`."
            )

        window_ = np.hamming(self.kernel_size)[: self.half_kernel]  # Half window
        n_ = (
            2 * np.pi * (torch.arange(-self.half_kernel, 0.0).view(1, -1) / self.sample_rate)
        )  # Half time vector
        self.register_buffer("window_", torch.from_numpy(window_).float())
        self.register_buffer("n_", n_)

    def _initialize_filters(self):
        """ Filter Initialization along the Mel scale"""
        low_hz = 30
        high_hz = self.sample_rate / 2 - (self.min_low_hz + self.min_band_hz)
        mel = np.linspace(
            self.to_mel(low_hz), self.to_mel(high_hz), self.n_filters // 2 + 1, dtype="float32"
        )
        hz = self.to_hz(mel)
        # filters parameters (out_channels // 2, 1)
        self.low_hz_ = nn.Parameter(torch.from_numpy(hz[:-1]).view(-1, 1))
        self.band_hz_ = nn.Parameter(torch.from_numpy(np.diff(hz)).view(-1, 1))

    def filters(self):
        """ Compute filters from parameters """
        low = self.min_low_hz + torch.abs(self.low_hz_)
        high = torch.clamp(
            low + self.min_band_hz + torch.abs(self.band_hz_), self.min_low_hz, self.sample_rate / 2
        )
        cos_filters = self.make_filters(low, high, filt_type="cos")
        sin_filters = self.make_filters(low, high, filt_type="sin")
        return torch.cat([cos_filters, sin_filters], dim=0)

    def make_filters(self, low, high, filt_type: str = "cos"):
        band = (high - low)[:, 0]
        ft_low = torch.matmul(low, self.n_)
        ft_high = torch.matmul(high, self.n_)
        if filt_type == "cos":  # Even filters from the SincNet paper.
            bp_left = ((torch.sin(ft_high) - torch.sin(ft_low)) / (self.n_ / 2)) * self.window_
            bp_center = 2 * band.view(-1, 1)
            bp_right = torch.flip(bp_left, dims=[1])
        elif filt_type == "sin":  # Extension including odd filters
            bp_left = ((torch.cos(ft_low) - torch.cos(ft_high)) / (self.n_ / 2)) * self.window_
            bp_center = torch.zeros_like(band.view(-1, 1))
            bp_right = -torch.flip(bp_left, dims=[1])
        else:
            raise ValueError("Invalid filter type {}".format(filt_type))
        band_pass = torch.cat([bp_left, bp_center, bp_right], dim=1)
        band_pass = band_pass / (2 * band[:, None])
        return band_pass.view(self.n_filters // 2, 1, self.kernel_size)

    @staticmethod
    def to_mel(hz):
        return 2595 * np.log10(1 + hz / 700)

    @staticmethod
    def to_hz(mel):
        return 700 * (10 ** (mel / 2595) - 1)

    def get_config(self):
        """ Returns dictionary of arguments to re-instantiate the class."""
        config = {
            "min_low_hz": self.min_low_hz,
            "min_band_hz": self.min_band_hz,
        }
        base_config = super(ParamSincFB, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import functools
import torch


def is_tracing():
    # Taken for pytorch for compat in 1.6.0
    """
    Returns ``True`` in tracing (if a function is called during the tracing of
    code with ``torch.jit.trace``) and ``False`` otherwise.
    """
    return torch._C._is_tracing()


def script_if_tracing(fn):
    # Taken for pytorch for compat in 1.6.0
    """
    Compiles ``fn`` when it is first called during tracing. ``torch.jit.script``
    has a non-negligible start up time when it is first called due to
    lazy-initializations of many compiler builtins. Therefore you should not use
    it in library code. However, you may want to have parts of your library work
    in tracing even if they use control flow. In these cases, you should use
    ``@torch.jit.script_if_tracing`` to substitute for
    ``torch.jit.script``.

    Arguments:
        fn: A function to compile.

    Returns:
        If called during tracing, a :class:`ScriptFunction` created by `torch.jit.script` is returned.
        Otherwise, the original function `fn` is returned.
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if not is_tracing():
            # Not tracing, don't do anything
            return fn(*args, **kwargs)

        compiled_fn = torch.jit.script(wrapper.__original_fn)  # type: ignore
        return compiled_fn(*args, **kwargs)

    wrapper.__original_fn = fn  # type: ignore
    wrapper.__script_if_tracing_wrapper = True  # type: ignore

    return wrapper



from .enc_dec import Filterbank, Encoder, Decoder
from .analytic_free_fb import AnalyticFreeFB
from .free_fb import FreeFB
from .param_sinc_fb import ParamSincFB
from .stft_fb import STFTFB
from .torch_stft_fb import TorchSTFTFB
from .griffin_lim import griffin_lim, misi
from .multiphase_gammatone_fb import MultiphaseGammatoneFB
from .melgram_fb import MelGramFB

__version__ = "0.4.0"


def make_enc_dec(
    fb_name,
    n_filters,
    kernel_size,
    stride=None,
    sample_rate=8000.0,
    who_is_pinv=None,
    padding=0,
    output_padding=0,
    **kwargs,
):
    """Creates congruent encoder and decoder from the same filterbank family.

    Args:
        fb_name (str, className): Filterbank family from which to make encoder
            and decoder. To choose among [``'free'``, ``'analytic_free'``,
            ``'param_sinc'``, ``'stft'``]. Can also be a class defined in a
            submodule in this subpackade (e.g. :class:`~.FreeFB`).
        n_filters (int): Number of filters.
        kernel_size (int): Length of the filters.
        stride (int, optional): Stride of the convolution.
            If None (default), set to ``kernel_size // 2``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.0.
        who_is_pinv (str, optional): If `None`, no pseudo-inverse filters will
            be used. If string (among [``'encoder'``, ``'decoder'``]), decides
            which of ``Encoder`` or ``Decoder`` will be the pseudo inverse of
            the other one.
        padding (int): Zero-padding added to both sides of the input.
            Passed to Encoder and Decoder.
        output_padding (int): Additional size added to one side of the output shape.
            Passed to Decoder.
        **kwargs: Arguments which will be passed to the filterbank class
            additionally to the usual `n_filters`, `kernel_size` and `stride`.
            Depends on the filterbank family.
    Returns:
        :class:`.Encoder`, :class:`.Decoder`
    """
    fb_class = get(fb_name)

    if who_is_pinv in ["dec", "decoder"]:
        fb = fb_class(n_filters, kernel_size, stride=stride, sample_rate=sample_rate, **kwargs)
        enc = Encoder(fb, padding=padding)
        # Decoder filterbank is pseudo inverse of encoder filterbank.
        dec = Decoder.pinv_of(fb)
    elif who_is_pinv in ["enc", "encoder"]:
        fb = fb_class(n_filters, kernel_size, stride=stride, sample_rate=sample_rate, **kwargs)
        dec = Decoder(fb, padding=padding, output_padding=output_padding)
        # Encoder filterbank is pseudo inverse of decoder filterbank.
        enc = Encoder.pinv_of(fb)
    else:
        fb = fb_class(n_filters, kernel_size, stride=stride, sample_rate=sample_rate, **kwargs)
        enc = Encoder(fb, padding=padding)
        # Filters between encoder and decoder should not be shared.
        fb = fb_class(n_filters, kernel_size, stride=stride, sample_rate=sample_rate, **kwargs)
        dec = Decoder(fb, padding=padding, output_padding=output_padding)
    return enc, dec


def register_filterbank(custom_fb):
    """Register a custom filterbank, gettable with `filterbanks.get`.

    Args:
        custom_fb: Custom filterbank to register.

    """
    if custom_fb.__name__ in globals().keys() or custom_fb.__name__.lower() in globals().keys():
        raise ValueError(f"Filterbank {custom_fb.__name__} already exists. Choose another name.")
    globals().update({custom_fb.__name__: custom_fb})


def get(identifier):
    """Returns a filterbank class from a string. Returns its input if it
    is callable (already a :class:`.Filterbank` for example).

    Args:
        identifier (str or Callable or None): the filterbank identifier.

    Returns:
        :class:`.Filterbank` or None
    """
    if identifier is None:
        return None
    elif callable(identifier):
        return identifier
    elif isinstance(identifier, str):
        cls = globals().get(identifier)
        if cls is None:
            raise ValueError("Could not interpret filterbank identifier: " + str(identifier))
        return cls
    else:
        raise ValueError("Could not interpret filterbank identifier: " + str(identifier))


# Aliases.
free = FreeFB
analytic_free = AnalyticFreeFB
param_sinc = ParamSincFB
stft = STFTFB
torch_stft = TorchSTFTFB
multiphase_gammatone = mpgtf = MultiphaseGammatoneFB

# For the docs
__all__ = [
    "Filterbank",
    "Encoder",
    "Decoder",
    "FreeFB",
    "STFTFB",
    "TorchSTFTFB",
    "AnalyticFreeFB",
    "ParamSincFB",
    "MultiphaseGammatoneFB",
    "MelGramFB",
    "griffin_lim",
    "misi",
    "make_enc_dec",
]



import warnings
from typing import Optional
import torch
from torch import nn
from torch.nn import functional as F

from .scripting import script_if_tracing


class Filterbank(nn.Module):
    """Base Filterbank class.
    Each subclass has to implement a ``filters`` method.

    Args:
        n_filters (int): Number of filters.
        kernel_size (int): Length of the filters.
        stride (int, optional): Stride of the conv or transposed conv. (Hop size).
            If None (default), set to ``kernel_size // 2``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.

    Attributes:
        n_feats_out (int): Number of output filters.
    """

    def __init__(self, n_filters, kernel_size, stride=None, sample_rate=8000.0):
        super(Filterbank, self).__init__()
        self.n_filters = n_filters
        self.kernel_size = kernel_size
        self.stride = stride if stride else self.kernel_size // 2
        # If not specified otherwise in the filterbank's init, output
        # number of features is equal to number of required filters.
        self.n_feats_out = n_filters
        self.sample_rate = sample_rate

    def filters(self):
        """ Abstract method for filters. """
        raise NotImplementedError

    def pre_analysis(self, wav: torch.Tensor):
        """Apply transform before encoder convolution."""
        return wav

    def post_analysis(self, spec: torch.Tensor):
        """Apply transform to encoder convolution."""
        return spec

    def pre_synthesis(self, spec: torch.Tensor):
        """Apply transform before decoder transposed convolution."""
        return spec

    def post_synthesis(self, wav: torch.Tensor):
        """Apply transform after decoder transposed convolution."""
        return wav

    def get_config(self):
        """Returns dictionary of arguments to re-instantiate the class.
        Needs to be subclassed if the filterbanks takes additional arguments
        than ``n_filters`` ``kernel_size`` ``stride`` and ``sample_rate``.
        """
        config = {
            "fb_name": self.__class__.__name__,
            "n_filters": self.n_filters,
            "kernel_size": self.kernel_size,
            "stride": self.stride,
            "sample_rate": self.sample_rate,
        }
        return config

    def forward(self, waveform):
        raise NotImplementedError("Filterbanks must be wrapped with an Encoder or a Decoder.")


class _EncDec(nn.Module):
    """Base private class for Encoder and Decoder.
    Common parameters and methods.

    Args:
        filterbank (:class:`Filterbank`): Filterbank instance. The filterbank
            to use as an encoder or a decoder.
        is_pinv (bool): Whether to be the pseudo inverse of filterbank.

    Attributes:
        filterbank (:class:`Filterbank`)
        stride (int)
        is_pinv (bool)
    """

    def __init__(self, filterbank, is_pinv=False):
        super(_EncDec, self).__init__()
        self.filterbank = filterbank
        self.sample_rate = getattr(filterbank, "sample_rate", None)
        self.stride = self.filterbank.stride
        self.is_pinv = is_pinv

    def filters(self):
        return self.filterbank.filters()

    def compute_filter_pinv(self, filters):
        """ Computes pseudo inverse filterbank of given filters."""
        scale = self.filterbank.stride / self.filterbank.kernel_size
        shape = filters.shape
        ifilt = torch.pinverse(filters.squeeze()).transpose(-1, -2).view(shape)
        # Compensate for the overlap-add.
        return ifilt * scale

    def get_filters(self):
        """ Returns filters or pinv filters depending on `is_pinv` attribute """
        if self.is_pinv:
            return self.compute_filter_pinv(self.filters())
        else:
            return self.filters()

    def get_config(self):
        """ Returns dictionary of arguments to re-instantiate the class."""
        config = {"is_pinv": self.is_pinv}
        base_config = self.filterbank.get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Encoder(_EncDec):
    r"""Encoder class.

    Add encoding methods to Filterbank classes.
    Not intended to be subclassed.

    Args:
        filterbank (:class:`Filterbank`): The filterbank to use
            as an encoder.
        is_pinv (bool): Whether to be the pseudo inverse of filterbank.
        as_conv1d (bool): Whether to behave like nn.Conv1d.
            If True (default), forwarding input with shape :math:`(batch, 1, time)`
            will output a tensor of shape :math:`(batch, freq, conv\_time)`.
            If False, will output a tensor of shape :math:`(batch, 1, freq, conv\_time)`.
        padding (int): Zero-padding added to both sides of the input.

    """

    def __init__(self, filterbank, is_pinv=False, as_conv1d=True, padding=0):
        super(Encoder, self).__init__(filterbank, is_pinv=is_pinv)
        self.as_conv1d = as_conv1d
        self.n_feats_out = self.filterbank.n_feats_out
        self.padding = padding

    @classmethod
    def pinv_of(cls, filterbank, **kwargs):
        """Returns an :class:`~.Encoder`, pseudo inverse of a
        :class:`~.Filterbank` or :class:`~.Decoder`."""
        if isinstance(filterbank, Filterbank):
            return cls(filterbank, is_pinv=True, **kwargs)
        elif isinstance(filterbank, Decoder):
            return cls(filterbank.filterbank, is_pinv=True, **kwargs)

    def forward(self, waveform):
        """Convolve input waveform with the filters from a filterbank.

        Args:
            waveform (:class:`torch.Tensor`): any tensor with samples along the
                last dimension. The waveform representation with and
                batch/channel etc.. dimension.

        Returns:
            :class:`torch.Tensor`: The corresponding TF domain signal.

        Shapes
            >>> (time, ) -> (freq, conv_time)
            >>> (batch, time) -> (batch, freq, conv_time)  # Avoid
            >>> if as_conv1d:
            >>>     (batch, 1, time) -> (batch, freq, conv_time)
            >>>     (batch, chan, time) -> (batch, chan, freq, conv_time)
            >>> else:
            >>>     (batch, chan, time) -> (batch, chan, freq, conv_time)
            >>> (batch, any, dim, time) -> (batch, any, dim, freq, conv_time)
        """
        filters = self.get_filters()
        waveform = self.filterbank.pre_analysis(waveform)
        spec = multishape_conv1d(
            waveform,
            filters=filters,
            stride=self.stride,
            padding=self.padding,
            as_conv1d=self.as_conv1d,
        )
        return self.filterbank.post_analysis(spec)


@script_if_tracing
def multishape_conv1d(
    waveform: torch.Tensor,
    filters: torch.Tensor,
    stride: int,
    padding: int = 0,
    as_conv1d: bool = True,
) -> torch.Tensor:
    if waveform.ndim == 1:
        # Assumes 1D input with shape (time,)
        # Output will be (freq, conv_time)
        return F.conv1d(waveform[None, None], filters, stride=stride, padding=padding).squeeze()
    elif waveform.ndim == 2:
        # Assume 2D input with shape (batch or channels, time)
        # Output will be (batch or channels, freq, conv_time)
        warnings.warn(
            "Input tensor was 2D. Applying the corresponding "
            "Decoder to the current output will result in a 3D "
            "tensor. This behaviours was introduced to match "
            "Conv1D and ConvTranspose1D, please use 3D inputs "
            "to avoid it. For example, this can be done with "
            "input_tensor.unsqueeze(1)."
        )
        return F.conv1d(waveform.unsqueeze(1), filters, stride=stride, padding=padding)
    elif waveform.ndim == 3:
        batch, channels, time_len = waveform.shape
        if channels == 1 and as_conv1d:
            # That's the common single channel case (batch, 1, time)
            # Output will be (batch, freq, stft_time), behaves as Conv1D
            return F.conv1d(waveform, filters, stride=stride, padding=padding)
        else:
            # Return batched convolution, input is (batch, 3, time), output will be
            # (b, 3, f, conv_t). Useful for multichannel transforms. If as_conv1d is
            # false, (batch, 1, time) will output (batch, 1, freq, conv_time), useful for
            # consistency.
            return batch_packed_1d_conv(waveform, filters, stride=stride, padding=padding)
    else:  # waveform.ndim > 3
        # This is to compute "multi"multichannel convolution.
        # Input can be (*, time), output will be (*, freq, conv_time)
        return batch_packed_1d_conv(waveform, filters, stride=stride, padding=padding)


def batch_packed_1d_conv(
    inp: torch.Tensor, filters: torch.Tensor, stride: int = 1, padding: int = 0
):
    # Here we perform multichannel / multi-source convolution.
    # Output should be (batch, channels, freq, conv_time)
    batched_conv = F.conv1d(inp.view(-1, 1, inp.shape[-1]), filters, stride=stride, padding=padding)
    output_shape = inp.shape[:-1] + batched_conv.shape[-2:]
    return batched_conv.view(output_shape)


class Decoder(_EncDec):
    """Decoder class.

    Add decoding methods to Filterbank classes.
    Not intended to be subclassed.

    Args:
        filterbank (:class:`Filterbank`): The filterbank to use as an decoder.
        is_pinv (bool): Whether to be the pseudo inverse of filterbank.
        padding (int): Zero-padding added to both sides of the input.
        output_padding (int): Additional size added to one side of the
            output shape.

    .. note::
        ``padding`` and ``output_padding`` arguments are directly passed to
        ``F.conv_transpose1d``.
    """

    def __init__(self, filterbank, is_pinv=False, padding=0, output_padding=0):
        super().__init__(filterbank, is_pinv=is_pinv)
        self.padding = padding
        self.output_padding = output_padding

    @classmethod
    def pinv_of(cls, filterbank):
        """ Returns an Decoder, pseudo inverse of a filterbank or Encoder."""
        if isinstance(filterbank, Filterbank):
            return cls(filterbank, is_pinv=True)
        elif isinstance(filterbank, Encoder):
            return cls(filterbank.filterbank, is_pinv=True)

    def forward(self, spec, length: Optional[int] = None) -> torch.Tensor:
        """Applies transposed convolution to a TF representation.

        This is equivalent to overlap-add.

        Args:
            spec (:class:`torch.Tensor`): 3D or 4D Tensor. The TF
                representation. (Output of :func:`Encoder.forward`).
            length: desired output length.

        Returns:
            :class:`torch.Tensor`: The corresponding time domain signal.
        """
        filters = self.get_filters()
        spec = self.filterbank.pre_synthesis(spec)
        wav = multishape_conv_transpose1d(
            spec,
            filters,
            stride=self.stride,
            padding=self.padding,
            output_padding=self.output_padding,
        )
        wav = self.filterbank.post_synthesis(wav)
        if length is not None:
            length = min(length, wav.shape[-1])
            return wav[..., :length]
        return wav


@script_if_tracing
def multishape_conv_transpose1d(
    spec: torch.Tensor,
    filters: torch.Tensor,
    stride: int = 1,
    padding: int = 0,
    output_padding: int = 0,
) -> torch.Tensor:
    if spec.ndim == 2:
        # Input is (freq, conv_time), output is (time)
        return F.conv_transpose1d(
            spec.unsqueeze(0),
            filters,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
        ).squeeze()
    if spec.ndim == 3:
        # Input is (batch, freq, conv_time), output is (batch, 1, time)
        return F.conv_transpose1d(
            spec,
            filters,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
        )
    else:
        # Multiply all the left dimensions together and group them in the
        # batch. Make the convolution and restore.
        view_as = (-1,) + spec.shape[-2:]
        out = F.conv_transpose1d(
            spec.reshape(view_as),
            filters,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
        )
        return out.view(spec.shape[:-2] + (-1,))



import torch
import torch.nn as nn
from .enc_dec import Filterbank


class FreeFB(Filterbank):
    """Free filterbank without any constraints. Equivalent to
    :class:`nn.Conv1d`.

    Args:
        n_filters (int): Number of filters.
        kernel_size (int): Length of the filters.
        stride (int, optional): Stride of the convolution.
            If None (default), set to ``kernel_size // 2``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.

    Attributes:
        n_feats_out (int): Number of output filters.

    References
        [1] : "Filterbank design for end-to-end speech separation". ICASSP 2020.
        Manuel Pariente, Samuele Cornell, Antoine Deleforge, Emmanuel Vincent.
    """

    def __init__(self, n_filters, kernel_size, stride=None, sample_rate=8000.0, **kwargs):
        super().__init__(n_filters, kernel_size, stride=stride, sample_rate=sample_rate)
        self._filters = nn.Parameter(torch.ones(n_filters, 1, kernel_size))
        for p in self.parameters():
            nn.init.xavier_normal_(p)

    def filters(self):
        return self._filters



import math
import torch
from typing import Tuple

from .scripting import script_if_tracing
from .deprecation import mark_deprecated


def mul_c(inp, other, dim: int = -2):
    """Entrywise product for complex valued tensors.

    Operands are assumed to have the real parts of each entry followed by the
    imaginary parts of each entry along dimension `dim`, e.g. for,
    ``dim = 1``, the matrix

    .. code::

        [[1, 2, 3, 4],
         [5, 6, 7, 8]]

    is interpreted as

    .. code::

        [[1 + 3j, 2 + 4j],
         [5 + 7j, 6 + 8j]

    where `j` is such that `j * j = -1`.

    Args:
        inp (:class:`torch.Tensor`): The first operand with real and
            imaginary parts concatenated on the `dim` axis.
        other (:class:`torch.Tensor`): The second operand.
        dim (int, optional): frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.
    Returns:
        :class:`torch.Tensor`:
            The complex multiplication between `inp` and `other`

            For now, it assumes that `other` has the same shape as `inp` along
            `dim`.
    """
    check_complex(inp, dim=dim)
    check_complex(other, dim=dim)
    real1, imag1 = inp.chunk(2, dim=dim)
    real2, imag2 = other.chunk(2, dim=dim)
    return torch.cat([real1 * real2 - imag1 * imag2, real1 * imag2 + imag1 * real2], dim=dim)


def reim(x, dim: int = -2) -> Tuple[torch.Tensor, torch.Tensor]:
    """Returns a tuple (re, im).

    Args:
        x (:class:`torch.Tensor`): Complex valued tensor.
        dim (int): frequency (or equivalent) dimension along which real and
            imaginary values are concatenated.
    """
    return torch.chunk(x, 2, dim=dim)


def mag(x, dim: int = -2, EPS: float = 1e-8):
    """Takes the magnitude of a complex tensor.

    The operands is assumed to have the real parts of each entry followed by
    the imaginary parts of each entry along dimension `dim`, e.g. for,
    ``dim = 1``, the matrix

    .. code::

        [[1, 2, 3, 4],
         [5, 6, 7, 8]]

    is interpreted as

    .. code::

        [[1 + 3j, 2 + 4j],
         [5 + 7j, 6 + 8j]

    where `j` is such that `j * j = -1`.

    Args:
        x (:class:`torch.Tensor`): Complex valued tensor.
        dim (int): frequency (or equivalent) dimension along which real and
            imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`: The magnitude of x.
    """
    check_complex(x, dim=dim)
    power = torch.stack(torch.chunk(x, 2, dim=dim), dim=-1).pow(2).sum(dim=-1)
    power = power + EPS
    return power.pow(0.5)


def magreim(x, dim: int = -2):
    """Returns a concatenation of (mag, re, im).

    Args:
        x (:class:`torch.Tensor`): Complex valued tensor.
        dim (int): frequency (or equivalent) dimension along which real and
            imaginary values are concatenated.
    """
    return torch.cat([mag(x, dim=dim), x], dim=dim)


def apply_real_mask(tf_rep, mask, dim: int = -2):
    """Applies a real-valued mask to a real-valued representation.

    It corresponds to ReIm mask in [1].

    Args:
        tf_rep (:class:`torch.Tensor`): The time frequency representation to
            apply the mask to.
        mask (:class:`torch.Tensor`): The real-valued mask to be applied.
        dim (int): Kept to have the same interface with the other ones.
    Returns:
        :class:`torch.Tensor`: `tf_rep` multiplied by the `mask`.
    """
    return tf_rep * mask


def apply_mag_mask(tf_rep, mask, dim: int = -2):
    """Applies a real-valued mask to a complex-valued representation.

    If `tf_rep` has 2N elements along `dim`, `mask` has N elements, `mask` is
    duplicated along `dim` to apply the same mask to both the Re and Im.

    `tf_rep` is assumed to have the real parts of each entry followed by
    the imaginary parts of each entry along dimension `dim`, e.g. for,
    ``dim = 1``, the matrix

    .. code::

        [[1, 2, 3, 4],
         [5, 6, 7, 8]]

    is interpreted as

    .. code::

        [[1 + 3j, 2 + 4j],
         [5 + 7j, 6 + 8j]

    where `j` is such that `j * j = -1`.

    Args:
        tf_rep (:class:`torch.Tensor`): The time frequency representation to
            apply the mask to. Re and Im are concatenated along `dim`.
        mask (:class:`torch.Tensor`): The real-valued mask to be applied.
        dim (int): The frequency (or equivalent) dimension of both `tf_rep` and
            `mask` along which real and imaginary values are concatenated.
    Returns:
        :class:`torch.Tensor`: `tf_rep` multiplied by the `mask`.
    """
    check_complex(tf_rep, dim=dim)
    mask = torch.cat([mask, mask], dim=dim)
    return tf_rep * mask


def apply_complex_mask(tf_rep, mask, dim: int = -2):
    """Applies a complex-valued mask to a complex-valued representation.

    Operands are assumed to have the real parts of each entry followed by the
    imaginary parts of each entry along dimension `dim`, e.g. for,
    ``dim = 1``, the matrix

    .. code::

        [[1, 2, 3, 4],
         [5, 6, 7, 8]]

    is interpreted as

    .. code::

        [[1 + 3j, 2 + 4j],
         [5 + 7j, 6 + 8j]

    where `j` is such that `j * j = -1`.

    Args:
        tf_rep (:class:`torch.Tensor`): The time frequency representation to
            apply the mask to.
        mask (class:`torch.Tensor`): The complex-valued mask to be applied.
        dim (int): The frequency (or equivalent) dimension of both `tf_rep` an
            `mask` along which real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            `tf_rep` multiplied by the `mask` in the complex sense.
    """
    check_complex(tf_rep, dim=dim)
    return mul_c(tf_rep, mask, dim=dim)


def is_asteroid_complex(tensor, dim: int = -2):
    """Check if tensor is complex-like in a given dimension.

    Args:
        tensor (torch.Tensor): tensor to be checked.
        dim(int): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        True if dimension is even in the specified dimension, otherwise False

    """
    return tensor.shape[dim] % 2 == 0


def check_complex(tensor, dim: int = -2):
    """Assert that tensor is an Asteroid-style complex in a given dimension.

    Args:
        tensor (torch.Tensor): tensor to be checked.
        dim(int): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Raises:
        AssertionError if dimension is not even in the specified dimension

    """
    if not is_asteroid_complex(tensor, dim):
        raise AssertionError(
            f"Could not equally chunk the tensor (shape {tensor.shape}) "
            f"along the given dimension ({dim}). Dim axis is "
            "probably wrong"
        )


def to_numpy(tensor, dim: int = -2):
    """Convert complex-like torch tensor to numpy complex array

    Args:
        tensor (torch.Tensor): Complex tensor to convert to numpy.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.
    Returns:
        :class:`numpy.array`:
            Corresponding complex array.
    """
    check_complex(tensor, dim=dim)
    real, imag = torch.chunk(tensor, 2, dim=dim)
    return real.data.numpy() + 1j * imag.data.numpy()


def from_numpy(array, dim: int = -2):
    """Convert complex numpy array to complex-like torch tensor.

    Args:
        array (np.array): array to be converted.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            Corresponding torch.Tensor (complex axis in dim `dim`=
    """
    import numpy as np  # Hub-importable

    return torch.cat([torch.from_numpy(np.real(array)), torch.from_numpy(np.imag(array))], dim=dim)


@script_if_tracing
def is_torchaudio_complex(x):
    """Check if tensor is Torchaudio-style complex-like (last dimension is 2).

    Args:
        x (torch.Tensor): tensor to be checked.

    Returns:
        True if last dimension is 2, else False.
    """
    return x.shape[-1] == 2


@script_if_tracing
def check_torchaudio_complex(tensor):
    """Assert that tensor is Torchaudo-style complex-like (last dimension is 2).

    Args:
        tensor (torch.Tensor): tensor to be checked.

    Raises:
        AssertionError if last dimension is != 2.
    """
    if not is_torchaudio_complex(tensor):
        raise AssertionError(
            f"Tensor of shape {tensor.shape} is not Torchaudio-style complex-like"
            "(expected last dimension to be == 2)"
        )


def to_torchaudio(tensor, dim: int = -2):
    """Converts complex-like torch tensor to torchaudio style complex tensor.

    Args:
        tensor (torch.tensor): asteroid-style complex-like torch tensor.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            torchaudio-style complex-like torch tensor.
    """
    return torch.stack(torch.chunk(tensor, 2, dim=dim), dim=-1)


@script_if_tracing
def from_torchaudio(tensor, dim: int = -2):
    """Converts torchaudio style complex tensor to complex-like torch tensor.

    Args:
        tensor (torch.tensor): torchaudio-style complex-like torch tensor.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            asteroid-style complex-like torch tensor.
    """
    dim = dim - 1 if dim < 0 else dim
    return torch.cat(torch.chunk(tensor, 2, dim=-1), dim=dim).squeeze(-1)


def to_torch_complex(tensor, dim: int = -2):
    """Converts complex-like torch tensor to native PyTorch complex tensor.

    Args:
        tensor (torch.tensor): asteroid-style complex-like torch tensor.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            Pytorch native complex-like torch tensor.
    """
    return torch.view_as_complex(to_torchaudio(tensor, dim=dim))


def from_torch_complex(tensor, dim: int = -2):
    """Converts Pytorch native complex tensor to complex-like torch tensor.

    Args:
        tensor (torch.tensor): PyTorch native complex-like torch tensor.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            asteroid-style complex-like torch tensor.
    """
    return torch.cat([tensor.real, tensor.imag], dim=dim)


def angle(tensor, dim: int = -2):
    """Return the angle of the complex-like torch tensor.

    Args:
        tensor (torch.Tensor): the complex tensor from which to extract the
            phase.
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            The counterclockwise angle from the positive real axis on
            the complex plane in radians.
    """
    check_complex(tensor, dim=dim)
    real, imag = torch.chunk(tensor, 2, dim=dim)
    return torch.atan2(imag, real)


def from_magphase(mag_spec, phase, dim: int = -2):
    """Return a complex-like torch tensor from magnitude and phase components.

    Args:
        mag_spec (torch.tensor): magnitude of the tensor.
        phase (torch.tensor): angle of the tensor
        dim(int, optional): the frequency (or equivalent) dimension along which
            real and imaginary values are concatenated.

    Returns:
        :class:`torch.Tensor`:
            The corresponding complex-like torch tensor.
    """
    return torch.cat([mag_spec * torch.cos(phase), mag_spec * torch.sin(phase)], dim=dim)


def magphase(spec: torch.Tensor, dim: int = -2) -> Tuple[torch.Tensor, torch.Tensor]:
    """Splits Asteroid complex-like tensor into magnitude and phase."""
    mag_val = mag(spec, dim=dim)
    phase = angle(spec, dim=dim)
    return mag_val, phase


def centerfreq_correction(
    spec: torch.Tensor,
    kernel_size: int,
    stride: int = None,
    dim: int = -2,
) -> torch.Tensor:
    """Corrects phase from the input spectrogram so that a sinusoid in the
    middle of a bin keeps the same phase from one frame to the next.

    Args:
        spec: Spectrogram tensor of shape (batch, n_freq + 2, frames).
        kernel_size (int): Kernel size of the STFT.
        stride (int): Stride of the STFT.
        dim (int): Only works of dim=-2.

    Returns:
        Tensor: the input spec with corrected phase.
    """
    if dim != -2:
        raise NotImplementedError
    if stride is None:
        stride = kernel_size // 2
    # Phase will be (batch, n_freq // 2 + 1, frames)
    mag_spec, phase = magphase(spec, dim=dim)
    new_phase = phase_centerfreq_correction(phase, kernel_size=kernel_size, stride=stride)
    new_spec = from_magphase(mag_spec, new_phase, dim=dim)
    return new_spec


def phase_centerfreq_correction(
    phase: torch.Tensor,
    kernel_size: int,
    stride: int = None,
) -> torch.Tensor:
    """Corrects phase so that a sinusoid in the middle of a bin keeps the
    same phase from one frame to the next.

    Args:
        phase: tensor of shape (batch, n_freq//2 + 1, frames)
        kernel_size (int): Kernel size of the STFT.
        stride (int): Stride of the STFT.

    Returns:
        Tensor: corrected phase.
    """
    *_, freq, frames = phase.shape
    tmp = torch.arange(freq).unsqueeze(-1) * torch.arange(frames)[None]
    correction = -2 * tmp * stride * math.pi / kernel_size
    return phase + correction



import codecs
import os
import re

from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as fh:
    long_description = fh.read()

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="asteroid-filterbanks",
    version=find_version("asteroid_filterbanks", "__init__.py"),
    author="Manuel Pariente",
    author_email="manuel.pariente@loria.fr",
    url="https://github.com/asteroid-team/asteroid-filterbanks",
    description="Asteroid's filterbanks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    python_requires=">=3.6",
    install_requires=["numpy", "torch>=1.8.0", "typing_extensions"],
    extras_require={
        "all": ["librosa", "scipy"],
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import torch
import numpy as np
from .enc_dec import Filterbank


class STFTFB(Filterbank):
    """STFT filterbank.

    Args:
        n_filters (int): Number of filters. Determines the length of the STFT
            filters before windowing.
        kernel_size (int): Length of the filters (i.e the window).
        stride (int, optional): Stride of the convolution (hop size). If None
            (default), set to ``kernel_size // 2``.
        window (:class:`numpy.ndarray`, optional): If None, defaults to
            ``np.sqrt(np.hanning())``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.

    Attributes:
        n_feats_out (int): Number of output filters.
    """

    def __init__(
        self, n_filters, kernel_size, stride=None, window=None, sample_rate=8000.0, **kwargs
    ):
        super().__init__(n_filters, kernel_size, stride=stride, sample_rate=sample_rate)
        assert n_filters >= kernel_size
        if n_filters % 2 != 0:
            raise ValueError(f"n_filters must be even, got {n_filters}")
        self.cutoff = int(n_filters / 2 + 1)
        self.n_feats_out = 2 * self.cutoff

        if window is None:
            self.window = np.hanning(kernel_size + 1)[:-1] ** 0.5
        else:
            if isinstance(window, torch.Tensor):
                window = window.data.numpy()
            ws = window.size
            if not (ws == kernel_size):
                raise AssertionError(
                    f"Expected window of size {kernel_size}. Received {ws} instead."
                )
            self.window = window
        # Create and normalize DFT filters (can be overcomplete)
        filters = np.fft.fft(np.eye(n_filters))
        filters /= 0.5 * np.sqrt(kernel_size * n_filters / self.stride)

        # Keep only the windowed centered part to save computation.
        lpad = int((n_filters - kernel_size) // 2)
        rpad = int(n_filters - kernel_size - lpad)
        indexes = list(range(lpad, n_filters - rpad))
        filters = np.vstack(
            [np.real(filters[: self.cutoff, indexes]), np.imag(filters[: self.cutoff, indexes])]
        )

        filters[0, :] /= np.sqrt(2)
        filters[n_filters // 2, :] /= np.sqrt(2)
        filters = torch.from_numpy(filters * self.window).unsqueeze(1).float()
        self.register_buffer("_filters", filters)
        self.register_buffer("torch_window", torch.from_numpy(self.window).float())

    def filters(self):
        return self._filters

    def _load_from_state_dict(
        self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs
    ):
        """Compat loader to avoid breaking when torch_window is missing."""
        super()._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs
        )
        # Torch window won't change, it's just a convenience.
        to_remove = [key for key in missing_keys if key.endswith("torch_window")]
        for key in to_remove:
            missing_keys.remove(key)


def perfect_synthesis_window(analysis_window, hop_size):
    """Computes a window for perfect synthesis given an analysis window and
        a hop size.

    Args:
        analysis_window (np.array): Analysis window of the transform.
        hop_size (int): Hop size in number of samples.

    Returns:
        np.array : the synthesis window to use for perfectly inverting the STFT.
    """
    win_size = len(analysis_window)
    den = np.zeros_like(analysis_window)

    loop_on = (win_size - 1) // hop_size
    for win_idx in range(-loop_on, loop_on + 1):
        shifted = np.roll(analysis_window ** 2, win_idx * hop_size)
        if win_idx < 0:
            shifted[win_idx * hop_size :] = 0
        elif win_idx > 0:
            shifted[: win_idx * hop_size] = 0
        den += shifted
    den = np.where(den != 0.0, den, np.finfo(den.dtype).tiny)
    correction = int(0.5 * len(analysis_window) / hop_size)
    return correction * analysis_window / den



import torch
import torch.nn as nn
import numpy as np
from .enc_dec import Filterbank

try:
    from torch import rfft, irfft

    def conj(filt):
        return torch.stack([filt[:, :, :, 1], -filt[:, :, :, 0]], dim=-1)


except ImportError:
    from torch import fft

    # Very bad "wrapper" of torch1.7 fft module to match torch.rfft and irfft
    # just for the analytic filterbanks. Signal_dim is ignore and signal_sizes
    # is assumed to be of length 1 if provided. Don't use this anywhere, this is
    # just a quick fix.
    def rfft(input, signal_ndim, normalized=False):
        norm = "ortho" if normalized else "backward"
        return fft.fft(input, dim=-1, norm=norm)

    def irfft(input, signal_ndim, normalized=False, signal_sizes=None):
        norm = "ortho" if normalized else "backward"
        n = None if signal_sizes is None else signal_sizes[0]
        return fft.irfft(input, n=n, dim=-1, norm=norm)

    def conj(filt):
        return filt.conj()


class AnalyticFreeFB(Filterbank):
    """Free analytic (fully learned with analycity constraints) filterbank.
    For more details, see [1].

    Args:
        n_filters (int): Number of filters. Half of `n_filters` will
            have parameters, the other half will be the hilbert transforms.
            `n_filters` should be even.
        kernel_size (int): Length of the filters.
        stride (int, optional): Stride of the convolution.
            If None (default), set to ``kernel_size // 2``.
        sample_rate (float): Sample rate of the expected audio.
            Defaults to 8000.

    Attributes:
        n_feats_out (int): Number of output filters.

    References
        [1] : "Filterbank design for end-to-end speech separation". ICASSP 2020.
        Manuel Pariente, Samuele Cornell, Antoine Deleforge, Emmanuel Vincent.
    """

    def __init__(self, n_filters, kernel_size, stride=None, sample_rate=8000.0, **kwargs):
        super().__init__(n_filters, kernel_size, stride=stride, sample_rate=sample_rate)
        self.cutoff = int(n_filters // 2)
        self.n_feats_out = 2 * self.cutoff
        if n_filters % 2 != 0:
            print(
                "If the number of filters `n_filters` is odd, the "
                "output size of the layer will be `n_filters - 1`."
            )

        self._filters = nn.Parameter(torch.ones(n_filters // 2, 1, kernel_size), requires_grad=True)
        for p in self.parameters():
            nn.init.xavier_normal_(p, gain=1.0 / np.sqrt(2.0))

    def filters(self):
        ft_f = rfft(self._filters, 1, normalized=True)
        hft_f = conj(ft_f)
        hft_f = irfft(hft_f, 1, normalized=True, signal_sizes=(self.kernel_size,))
        return torch.cat([self._filters, hft_f], dim=0)



from torch import nn
import torch
from . import transforms
from typing import Union, Optional, Tuple

try:
    from typing import TypedDict
except ImportError:  # Fallback for >= Python 3.7
    from typing_extensions import TypedDict


class ExponentialMovingAverage(nn.Module):
    """
    Computes the exponential moving average of an sequential input.

    Influenced by leaf-audio's tensorflow implementation
        https://github.com/google-research/leaf-audio/blob/7ead2f9fe65da14c693c566fe8259ccaaf14129d/leaf_audio/postprocessing.py#L27

    See license here
        https://github.com/google-research/leaf-audio/blob/master/LICENSE
    """

    def __init__(
        self,
        smooth: float = 0.04,
        per_channel: bool = False,
        n_channels: int = 1,
        trainable: bool = False,
    ):
        super().__init__()
        if per_channel:
            weights = torch.full((n_channels,), fill_value=smooth)
        else:
            weights = torch.full((1,), fill_value=smooth)

        self.set_weights(data=weights, trainable=trainable)

    def set_weights(self, data: torch.Tensor, trainable: bool):
        self.weights = nn.Parameter(data, requires_grad=trainable)
        self.trainable = trainable

    def forward(
        self, mag_spec: torch.Tensor, initial_state: Optional[torch.Tensor]
    ) -> Tuple[torch.tensor, torch.Tensor]:
        if initial_state is None:
            initial_state = mag_spec[:, :, :, 0].unsqueeze(-1)

        weights = torch.clamp(self.weights, 0.0, 1.0)
        accumulator = initial_state.transpose(1, -1)
        out = []
        for x in torch.split(mag_spec.transpose(1, -1), 1, dim=1):
            accumulator = weights * x + (1.0 - weights) * accumulator
            out.append(accumulator.transpose(1, -1))
        return torch.cat(out, dim=-1), accumulator.transpose(1, -1)


class TrainableParameters(TypedDict):
    alpha: bool
    delta: bool
    root: bool
    smooth: bool


class _PCEN(nn.Module):
    def __init__(
        self,
        alpha: float = 0.96,
        delta: float = 2.0,
        root: float = 2.0,
        floor: float = 1e-6,
        smooth: float = 0.04,
        n_channels: int = 1,
        trainable: Union[bool, TrainableParameters] = False,
        per_channel_smoothing: bool = False,
    ):
        super().__init__()

        if trainable is True or trainable is False:
            trainable = TrainableParameters(
                alpha=trainable, delta=trainable, root=trainable, smooth=trainable
            )

        self.trainable = trainable
        self.n_channels = n_channels

        self.alpha = nn.Parameter(
            torch.full((self.n_channels,), fill_value=alpha), requires_grad=self.trainable["alpha"]
        )
        self.delta = nn.Parameter(
            torch.full((self.n_channels,), fill_value=delta), requires_grad=self.trainable["delta"]
        )
        self.root = nn.Parameter(
            torch.full((self.n_channels,), fill_value=root), requires_grad=self.trainable["root"]
        )

        self.floor = floor
        self.ema = ExponentialMovingAverage(
            smooth=smooth,
            per_channel=per_channel_smoothing,
            n_channels=self.n_channels,
            trainable=self.trainable["smooth"],
        )

    def forward(
        self, mag_spec: torch.Tensor, initial_state: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        post_squeeze = False
        if len(mag_spec.shape) == 3:
            # If n_channels is 1, add a single dimension to keep the shape consistent with multichannel shapes.
            mag_spec = mag_spec.unsqueeze(1)
            post_squeeze = True

        alpha = torch.min(self.alpha, torch.tensor(1.0))
        root = torch.max(self.root, torch.tensor(1.0))
        one_over_root = 1.0 / root
        ema_smoother, hidden_state = self.ema(mag_spec, initial_state)
        mag_spec = mag_spec.transpose(1, -1)
        ema_smoother = ema_smoother.transpose(1, -1)
        # Equation (1) in [1]
        out = (
            mag_spec / (self.floor + ema_smoother) ** alpha + self.delta
        ) ** one_over_root - self.delta ** one_over_root
        out = out.transpose(1, -1)
        if post_squeeze:
            out = out.squeeze(1)
        return out, hidden_state


class PCEN(_PCEN):
    """Per-Channel Energy Normalization as described in [1].

    This applies a fixed or learnable normalization by an exponential moving average smoother and a compression.

    Args:
        alpha: The AGC strength (or gain normalization strength) (α in the paper).
            Defaults to 0.96
        delta: Bias added before compression (δ in the paper).
            Defaults to 2.0
        root: One over exponent applied for compression (r in the paper).
            Defaults to 2.0
        floor: Offset added to compression, prevents division by zero. (ϵ in the paper)
            Defaults to 1e-6
        smooth: Smoothing coefficient (s in the paper).
            Defaults to 0.04
        n_channels: Number of channels in the time frequency representation.
            Defaults to 1
        trainable: If True, the parameters (alpha, delta, root and smooth) are trainable. If False, the parameters are fixed.
            Individual parameters can set to be fixed or trainable by passing a dictionary of booleans, with the key
            matching the parameter name and the value being either True (trainable) or False (fixed).
            i.e. ``{"alpha": False, "delta": True, "root": False, "smooth": True}``
            Defaults to False
        per_channel_smoothing: If True, each channel has it's own smoothing coefficient.
            Defaults to False

    Examples

        >>> audio = torch.randn(10, 2, 16_000)
        >>> fb = STFTFB(kernel_size=256, n_filters=256, stride=128)
        >>> enc = Encoder(fb)
        >>> tf_rep = enc(audio)
        >>> mag_spec = transform.mag(tf_rep)
        >>> pcen = PCEN(n_channels=2)
        >>> energy = pcen(mag_spec)
        >>> # or alternative with a mel-spectrogram
        >>> mel = MelScale(n_filters=256, sample_rate=16_000)
        >>> mel_spec = mel(audio)
        >>> energy = pcen(mag_spec)

    References
        [1]: Wang, Y., et al. "Trainable Frontend For Robust and Far-Field Keyword Spotting”, arXiv e-prints, 2016.
             https://arxiv.org/pdf/1607.05666.pdf

    Influenced by leaf-audio's tensorflow implementation
        https://github.com/google-research/leaf-audio/blob/7ead2f9fe65da14c693c566fe8259ccaaf14129d/leaf_audio/postprocessing.py

    See license here
        https://github.com/google-research/leaf-audio/blob/master/LICENSE
    """

    def forward(self, mag_spec: torch.Tensor) -> torch.Tensor:
        """Computes the PCEN from magnitude spectrum representation.

        Args:
            mag_spec: A tensor containing an magnitude spectrum representation.

        Shapes
            >>> (batch, n_channels, freq, time) -> (batch, n_channels, freq, time)
            >>> (batch, freq, time) -> (batch, freq, time) # Assumed to be mono-channel

        """
        return super().forward(mag_spec, None)[0]


class StatefulPCEN(_PCEN):
    """Per-Channel Energy Normalization as described in [1].

    Suitable for continuous real-time processing of incoming audio.

    This applies a fixed or learnable normalization by an exponential moving average smoother and a compression.

    Args:
        alpha: The AGC strength (or gain normalization strength) (α in the paper).
            Defaults to 0.96
        delta: Bias added before compression (δ in the paper).
            Defaults to 2.0
        root: One over exponent applied for compression (r in the paper).
            Defaults to 2.0
        floor: Offset added to compression, prevents division by zero. (ϵ in the paper)
            Defaults to 1e-6
        smooth: Smoothing coefficient (s in the paper).
            Defaults to 0.04
        n_channels: Number of channels in the time frequency representation.
            Defaults to 1
        trainable: If True, the parameters (alpha, delta, root and smooth) are trainable. If False, the parameters are fixed.
            Individual parameters can set to be fixed or trainable by passing a dictionary of booleans, with the key
            matching the parameter name and the value being either True (trainable) or False (fixed).
            i.e. ``{"alpha": False, "delta": True, "root": False, "smooth": True}``
            Defaults to False
        per_channel_smoothing: If True, each channel has it's own smoothing coefficient.
            Defaults to False

    Examples
        >>> audio = torch.randn(10, 2, 16_000)
        >>> fb = STFTFB(kernel_size=256, n_filters=256, stride=128)
        >>> enc = Encoder(fb)
        >>> tf_rep = enc(audio)
        >>> mag_spec = transform.mag(tf_rep)
        >>> mag_spec_1, mag_spec_2 = torch.chunk(mag_spec, 2, dim=-1)
        >>> pcen = PCEN(n_channels=2)
        >>> energy, hidden = pcen(mag_spec_1)
        >>> energy, hidden = pcen(mag_spec_2, hidden)

    References
        [1]: Wang, Y., et al. "Trainable Frontend For Robust and Far-Field Keyword Spotting”, arXiv e-prints, 2016.
             https://arxiv.org/pdf/1607.05666.pdf

    Influenced by leaf-audio's tensorflow implementation
        https://github.com/google-research/leaf-audio/blob/7ead2f9fe65da14c693c566fe8259ccaaf14129d/leaf_audio/postprocessing.py

    See license here
        https://github.com/google-research/leaf-audio/blob/master/LICENSE
    """

    @classmethod
    def from_pcen(cls, pcen: PCEN):
        stateful = cls()
        stateful.alpha = pcen.alpha
        stateful.delta = pcen.delta
        stateful.root = pcen.root
        stateful.floor = pcen.floor
        stateful.ema.set_weights(pcen.ema.weights.data, pcen.ema.trainable)
        return stateful

    def forward(
        self, mag_spec: torch.Tensor, initial_state: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Computes the PCEN from magnitude spectrum representation, and an optional smoothed version of the filterbank (Equation (2) in [1]).

        Args:
            mag_spec: tensor containing an magnitude spectrum representation.
            initial_state: A tensor containing the initial hidden state.
                If set to None, defaults to the first element by time (dim=-1) in the mag_spec tensor.
                Defaults to None
        Shapes
            >>> (batch, n_channels, freq, time) -> (batch, n_channels, freq, time), (batch_size, n_channel, freq, 1)
            >>> (batch, freq, time) -> (batch, freq, time), (batch_size, 1, freq, 1) # Assumed to be mono-channel

        References
            [1]: Wang, Y., et al. "Trainable Frontend For Robust and Far-Field Keyword Spotting”, arXiv e-prints, 2016.
                https://arxiv.org/pdf/1607.05666.pdf
        """
        return super().forward(mag_spec, initial_state)
